import os
from pathlib import Path

import pulumi
import yaml

from mesh.pulumi.resource import WireguardMesh, MeshArgs


def main():
    base_dir = Path(__file__).absolute().parent.parent.parent

    mesh_id = os.environ.get("PULUMI_MESH_ID", "mesh")
    mesh_file = Path(os.environ.get("PULUMI_MESH_FILE", "mesh.yaml"))

    with (base_dir / mesh_file).open() as f:
        mesh = WireguardMesh(mesh_id, MeshArgs(**yaml.safe_load(f)))
        pulumi.export(f"{mesh_id}:info", mesh.info)
        pulumi.export(f"{mesh_id}:nodes", mesh.nodes)


if __name__ == "__main__":
    main()



from functools import cached_property
from ipaddress import (
    IPv4Interface, IPv6Interface, IPv4Address, IPv6Address, IPv4Network, IPv6Network, ip_address, ip_interface, ip_network,
)
from random import getrandbits
from typing import Any, Self

from attrs import AttrsInstance, define, field, asdict
from attrs.converters import optional
from attrs.setters import frozen
import yaml

from .util import host_index

Network = IPv4Network | IPv6Network
Interface = IPv4Interface | IPv6Interface
Address = IPv4Address | IPv6Address

__all__ = (
    "IPv4Interface", "IPv6Interface", "IPv4Address", "IPv6Address", "IPv4Network", "IPv6Network",
    "ip_address", "ip_interface", "ip_network",
    "Network", "Interface", "Address",
    "RandomIPv6Address",
    "RandomIPv6Interface",
    "Endpoint", "NodeType",
    "DEFAULT_PORT",
)

DEFAULT_PORT = 51820


class RandomIPv6Address(IPv6Address):
    def __init__(self):
        super().__init__(b'\xfd' + getrandbits(120).to_bytes(15, 'big'))


class RandomIPv6Interface(IPv6Interface):
    def __init__(self):
        super().__init__(RandomIPv6Address())


def _ip_address_or_name(value: str) -> Address | str:
    try:
        return ip_address(value)
    except ValueError:
        return value


@define
class Endpoint(AttrsInstance):
    """Wireguard endpoint.

    Pedantic implementation to support ``wireguard-tools`` endpoint format.
    """
    host: Address | str = field(converter=_ip_address_or_name)
    port: int | None = field(default=None, converter=optional(int))

    @port.validator  # noqa
    def __port_validator(self, _: str, value: Any):
        if value is not None and not 0 <= value <= 65535:
            raise ValueError(f"Invalid port {value!r}")

    def __str__(self):
        return f"{self.host}:{self.port}" if self.port is not None else str(self.host)

    @classmethod
    def convert(cls, value: str | dict | Self) -> Self:
        if isinstance(value, str):
            return cls(*value.rsplit(":", 1))
        if isinstance(value, dict):
            return cls(**value)
        if isinstance(value, Endpoint):
            return value
        raise TypeError(f"Cannot parse endpoint from {value!r}")


@define(kw_only=True, slots=False, order=True, eq=True)
class NodeType(AttrsInstance):
    """A node in a Wireguard mesh network.
    """
    name: str = field(on_setattr=frozen)

    addr: Interface = field(
        converter=lambda x: ip_interface(x) if not isinstance(x, (IPv4Interface, IPv6Interface)) else x,
        on_setattr=frozen,
    )
    """Network bridge address for the node.
    
    Must be in the mesh network.
    """

    ssh: str | dict = field()
    """SSH connection string or arguments to the node.
    
    It is recommended to use the default SSH config file to specify connection details.
    
    This field is passed directly to ``fabric.Connection``, so any valid connection string
    can be used.
    """

    endpoint: Endpoint = field(converter=Endpoint.convert)
    """Public or network-accessible Wireguard endpoint."""

    listen_port: int | None = field(default=None, converter=optional(int))
    """Wireguard listen port."""

    json: dict[str, Any] | None = None
    """Set ``friendly_json`` data for all peers."""

    prio: int | None = field(default=None)
    """Bridge priority for GRE tunnels, normalized to ``-8 <= prio <= 7``.
    
    Calculated as ``32768 + 4096 * prio``.
    
    This field is used to create the initial bridge configurations, and should not be changed.
    
    When None, the bridge priority will be calculated as ``-8 + ((index-1) % 16)``.
    - Having any difference in bridge priority between nodes helps to avoid bad routing decisions.
    """

    @cached_property
    def index(self) -> int:
        return host_index(self.addr)

    @cached_property
    def ip(self) -> Address:
        return self.addr.ip

    @cached_property
    def network(self) -> Network:
        return self.addr.network

    def __str__(self):
        return f"[{self.index}] {self.addr.ip} <{self.name}> via {self.endpoint}"

    def asdict(self, *, serializer=None) -> dict:
        return asdict(self, filter=lambda a, v: v is not None and not a.name.startswith("_"), value_serializer=serializer)

    @name.validator  # noqa
    def __name_validator(self, _: str, value: Any):
        if not value:
            raise ValueError("Node name cannot be empty")

    @addr.validator  # noqa
    def __addr_validator(self, _: str, value: Interface):
        if (prefixlen := value.network.prefixlen) == value.max_prefixlen or prefixlen == 0:
            raise ValueError(f"Address {value!r} is not a network address")

    @ssh.validator  # noqa
    def __ssh_validator(self, _: str, value: Any):
        if isinstance(value, str):
            if not value:
                raise ValueError("SSH connection string cannot be empty")
        elif isinstance(value, dict):
            if "host" not in value:
                raise ValueError("SSH connection dict args must specify host")

    @listen_port.validator  # noqa
    def __port_validator(self, _: str, value: Any):
        if value is not None and not 0 <= value <= 65535:
            raise ValueError(f"Invalid port {value!r}")

    @prio.validator  # noqa
    def __prio_validator(self, _: str, value: Any):
        if value is not None and not -8 <= value <= 7:
            raise ValueError(f"Invalid bridge priority {value!r}")

    @classmethod
    def convert(cls, value, *, key: str | None = None) -> Self:
        if isinstance(value, cls):
            return value

        if isinstance(value, dict):
            if key is not None:
                value["name"] = value.setdefault("name", key) or key
            return cls(**value)

        raise TypeError(f"Cannot convert {value!r} to {cls.__name__}")

    @classmethod
    def convert_all(cls, value: dict | list) -> dict[str, Self]:
        if isinstance(value, dict):
            return {key: cls.convert(val, key=key) for key, val in value.items()}

        if isinstance(value, list):
            return {
                node_.name: (node_ := cls.convert(node))
                for node in value
            }

        raise TypeError(f"Cannot convert {value!r} to {cls.__name__}")


@define(kw_only=True, slots=False, order=True, eq=True)
class MeshType(AttrsInstance):
    name: str = field(on_setattr=frozen)

    network: Network = field(
        converter=lambda x: ip_network(x) if not isinstance(x, (IPv4Network, IPv6Network)) else x,
        on_setattr=frozen,
    )
    """Mesh network.

    This network will be routed among all nodes in the mesh.
    """

    full: bool = True
    """Peer all node pairs regardless of reachability.

    When False, checks whether two nodes can reach each other before peering them.
    """

    nodes: dict[str, NodeType]

    def asdict(self, *, serializer=None) -> dict:
        return dict(
            name=self.name,
            network=str(self.network),
            **(dict(full=False) if not self.full else {}),
            nodes={node.name: node.asdict(serializer=serializer) for node in self.nodes.values()},
        )

    def asyaml(self, *, serializer=None) -> str:
        return yaml.dump(self.asdict(serializer=serializer), sort_keys=False, allow_unicode=True)



"""Wireguard remote controller.
"""
from functools import cached_property
from io import StringIO
from time import sleep

from attrs import define
from fabric import Connection
from invoke import Promise
from wireguard_tools import WireguardConfig

__all__ = "Remote", "WireguardRemote",


@define(slots=False)
class Remote:
    connect: str | dict
    """SSH connection string or arguments to the node."""

    @property
    def host(self):
        return self.connect if isinstance(self.connect, str) else self.connect["host"]

    @cached_property
    def _connection(self) -> Connection:
        return Connection(**(
            self.connect if isinstance(self.connect, dict) else dict(host=self.connect)
        ))

    @cached_property
    def _is_root(self) -> bool:
        return self._connection.user == "root"

    def _run(self, command: str, *, root: bool = True, **kwds):
        kwds.setdefault("hide", True)
        kwds.setdefault("warn", True)
        return self._connection.run(
            f"sudo {command.lstrip()}" if root and not self._is_root else command,
            **kwds
        )

    def udping_send(self, host: str, port: int, data: str = "1") -> bool:
        return self._run(
            f"echo -n {data!r} > /dev/udp/{host}/{port}",
            root=False,
        ).ok

    def udping_recv(self, port: int, timeout: float = 1.0) -> Promise:
        return self._run(
            f"timeout {timeout} nc -u -l -W 1 0 {port}",
            asynchronous=True,
        )

    def udping_from(self, listen_port: int, endpoint_host: str, endpoint_port: int, remote: "Remote") -> bool:
        """Send UDP ping from another remote to an endpoint mapping to this remote.
        """
        data = "1"
        promise = self.udping_recv(listen_port)
        sleep(0.1)
        remote.udping_send(endpoint_host, endpoint_port, data=data)
        result = promise.join()
        return result.ok

    def udping_to(self, listen_port: int, endpoint_host: str, endpoint_port: int, remote: "Remote") -> bool:
        return remote.udping_from(listen_port, endpoint_host, endpoint_port, self)


@define
class WireguardRemote(Remote):
    """Wireguard remote controller.
    """
    interface: str

    @property
    def config(self) -> WireguardConfig | None:
        return WireguardConfig.from_wgconfig(StringIO(config_text)) if (config_text := self.config_text) is not None else None

    @property
    def config_text(self) -> str | None:
        if (command := self._run(f"cat /etc/wireguard/{self.interface}.conf")).ok:
            return command.stdout.strip()
        return None

    def config_write(self, config: str | WireguardConfig, *, warn: bool = False) -> None:
        if isinstance(config, WireguardConfig):
            config = config.to_wgconfig(wgquick_format=True)
        if self._is_root:
            self._run(f"cat > /etc/wireguard/{self.interface}.conf <<EOF\n{config.strip()}\nEOF", root=False, warn=warn)
        else:
            self._run(
                f"cat <<EOF | sudo tee /etc/wireguard/{self.interface}.conf >/dev/null\n{config.strip()}\nEOF",
                root=False,
                warn=warn,
            )

    def config_remove(self) -> None:
        if self.is_up:
            raise RuntimeError(f"[{self.connect}] Cannot remove config while {self.interface} is up")
        self._run(f"rm -f /etc/wireguard/{self.interface}.conf")

    @property
    def config_exists(self) -> bool:
        return self._run(f"test -f /etc/wireguard/{self.interface}.conf").ok

    @property
    def is_up(self) -> bool:
        return self._run(f"wg show {self.interface}").ok

    def up(self) -> str | RuntimeError:
        cmd = self._run(f"wg-quick up {self.interface} 2>&1")
        return cmd.stdout.strip() if cmd.ok else RuntimeError(cmd.stdout.strip())

    def down(self) -> str | RuntimeError:
        cmd = self._run(f"wg-quick down {self.interface} 2>&1")
        return cmd.stdout.strip() if cmd.ok else RuntimeError(cmd.stdout.strip())

    def restart(self) -> str | RuntimeError:
        cmd = self._run(f"wg-quick down {self.interface} 2>&1 && wg-quick up {self.interface} 2>&1")
        return cmd.stdout.strip() if cmd.ok else RuntimeError(cmd.stdout.strip())

    def show(self) -> str | RuntimeError:
        cmd = self._run(f"wg show {self.interface} 2>&1")
        return cmd.stdout.strip() if cmd.ok else RuntimeError(cmd.stdout.strip())



from hashlib import sha3_512
from ipaddress import IPv4Network, IPv6Network, IPv4Interface, IPv6Interface, ip_interface
from itertools import islice
from math import log2, ceil
from typing import Iterator

__all__ = "interface_name", "generate_subnets", "generate_hosts", "host_index",


def interface_name(key: str) -> str:
    return sha3_512(key.encode()).hexdigest()[:15]


def generate_subnets(network: IPv4Network | IPv6Network, count: int) -> Iterator[IPv4Network | IPv6Network]:
    """Generate a list of subnets from a network.

    Args:
        network (IPv4Network | IPv6Network): The network to generate subnets from.
        count (int): The number of subnets to generate.

    Returns:
        list[IPv4Network | IPv6Network]: The list of generated subnets.
    """
    return network.subnets(
        new_prefix=network.prefixlen + ceil(log2(count))
    )


def generate_hosts(
        network: IPv4Network | IPv6Network, count: int, prefixlen: int | None = None
) -> Iterator[IPv4Interface | IPv6Interface]:
    """Generate a list of host interfaces from a network.

    Args:
        network (IPv4Network | IPv6Network): The network to generate hosts from.
        count (int): The number of hosts to generate.
        prefixlen (int | None, optional): The prefix length to use for the generated hosts. Defaults to ``None``.

    Returns:
        list[IPv4Interface | IPv6Interface]: The list of generated host interfaces.
    """
    return islice(
        map(
            lambda addr: ip_interface(
                (addr, prefixlen if prefixlen is not None else network.prefixlen)
            ),
            network.hosts()
        ),
        count
    )


def host_index(host: IPv4Interface | IPv6Interface) -> int:
    """Get the index of a host in a network.

    Args:
        host (IPv4Interface | IPv6Interface): The host to get the index of.

    Returns:
        int: The index of the host.
    """
    return int(host.ip) - int(host.network.network_address)



"""Wireguard mesh network manager.
"""
import argparse
import json
import os
import sys

import yaml

from . import Mesh


def main():
    parser = argparse.ArgumentParser(prog="mesh", description=__doc__)
    parser.set_defaults(func=lambda _: parser.print_help())
    parser.add_argument(
        "-j", "--json", action="store_true",
        help="Input JSON instead of YAML (default: use file ext).",
    )
    parser.add_argument(
        "-J", "--json-out",
        action="store_true",
        help="Output JSON instead of YAML."
    )
    parser.add_argument(
        "-f", "--file", type=argparse.FileType("r"), default="mesh.yaml",
        help="Mesh configuration file or - for stdin (default: mesh.yaml).",
    )
    parser.add_argument("-q", "--quiet", action="store_true", help="Suppress output.")

    subparsers = parser.add_subparsers(title="commands", metavar="COMMAND", help="Command:", required=True)

    up_parser = subparsers.add_parser("up", help="- Bring up mesh.")
    up_parser.add_argument("-i", "--info", action="store_true", help="Show mesh info after bringing up.")
    up_parser.set_defaults(func=_up)
    down_parser = subparsers.add_parser("down", help="- Bring down mesh.")
    down_parser.add_argument(
        "-r", "--remove", action="store_true", help="Remove Wireguard interface configs."
    )
    down_parser.set_defaults(func=_down)
    sync_parser = subparsers.add_parser("sync", help="- Sync mesh.")
    sync_parser.set_defaults(func=_sync)
    show_parser = subparsers.add_parser("show", help="- Show mesh Wireguard info.")
    show_parser.set_defaults(func=_show)
    info_parser = subparsers.add_parser("info", help="- Show mesh network info.")
    info_parser.set_defaults(func=_info)
    conf_parser = subparsers.add_parser("conf", help="- Show mesh configuration.")
    conf_parser.set_defaults(func=_conf)

    args = parser.parse_args()

    if args.quiet:
        sys.stderr = open(os.devnull, "w")

    mesh_dict = (
        yaml.safe_load(args.file)
        if args.json is False or (args.json is None and args.file.name.endswith(".yaml")) else
        json.load(args.file)
    )

    mesh = Mesh(**mesh_dict)

    exit(args.func(mesh, args) or 0)


def _up(mesh: Mesh, args: argparse.Namespace):
    if up := mesh.up():
        if args.info:
            return _info(mesh, args)

    return int(not up)


def _down(mesh: Mesh, args: argparse.Namespace):
    return int(not mesh.down(remove=args.remove))


def _sync(mesh: Mesh, args: argparse.Namespace):
    return int(not mesh.sync())


def _show(mesh: Mesh, args: argparse.Namespace):
    mesh.show()
    return 0


def _info(mesh: Mesh, args: argparse.Namespace):
    print(yaml.dump(mesh.info, sort_keys=False) if not args.json_out else json.dumps(mesh.info, indent=4))
    return 0


def _conf(mesh: Mesh, args: argparse.Namespace):
    print(yaml.dump(mesh.conf, sort_keys=False) if not args.json_out else json.dumps(mesh.conf, indent=4))
    return 0


main()



from functools import cached_property
from secrets import token_bytes
import sys
from typing import TypedDict

from attrs import define, field
from wireguard_tools import WireguardConfig, WireguardKey
from wireguard_tools.wireguard_config import WireguardPeer

from .types import RandomIPv6Interface, Interface, NodeType, MeshType, ip_interface, DEFAULT_PORT
from .remote import WireguardRemote
from .gretap import gretap_up, gretap_down


@define(kw_only=True, slots=False, order=True, eq=True)
class MeshNode(NodeType):
    _mesh: MeshType | None = field(default=None, init=False, repr=False, eq=False, order=False)
    _config: WireguardConfig | None = field(default=None, init=False, repr=False, eq=False, order=False)

    @cached_property
    def remote(self) -> WireguardRemote:
        return WireguardRemote(
            connect=self.ssh,
            interface=f"wg-{self._mesh.name}{self.index}",
        )

    @property
    def mesh(self) -> MeshType:
        return self._mesh

    @property
    def config(self) -> WireguardConfig:
        return self._config

    @property
    def config_exists(self) -> bool:
        return self.remote.config_exists

    @property
    def wg_addr(self) -> Interface:
        """Wireguard address of the node.
        This address is only used to establish a GRE tunnel between mesh nodes, and is always a random IPv6/128 address.
        """
        return self._config.addresses[0]

    @property
    def peers(self) -> list[str]:
        return [peer.friendly_name for peer in self._config.peers.values()]

    @property
    def bridge_priority(self) -> int:
        prio = self.prio if self.prio is not None else -8 + (self.index - 1) % 16
        return 32768 + 4096 * prio

    # @property
    # def conf(self) -> dict:
    #     return super().conf | dict(
    #         address=str(self.network_addr.ip),
    #     )

    class Info(TypedDict):
        host: str
        is_up: bool
        config_exists: bool
        address: str
        peers: list[str]

    @property
    def info(self) -> Info:
        return self.Info(
            host=self.remote.host,
            is_up=self.remote.is_up,
            config_exists=self.remote.config_exists,
            address=str(self.addr.ip),
            peers=self.peers,
        )

    def __mesh_post_init__(self, mesh: MeshType):
        if self.network != mesh.network:
            raise ValueError(f"Node {self.name!r} address {self.addr} is not in the mesh network {mesh.network}")
        if mesh.nodes.get(self.name) is not self:
            raise ValueError(f"Node name {self.name!r} does not match mesh.nodes[{self.name!r}].")
        self._mesh = mesh
        self.__config_init()

    def __config_init(self):
        self._config = self.remote.config

        if self._config is not None:
            if self._config.listen_port != DEFAULT_PORT and self.listen_port is None:
                self.listen_port = self._config.listen_port

        else:
            self._config = WireguardConfig(
                private_key=WireguardKey.generate(),
                addresses=[RandomIPv6Interface()],
                listen_port=self.listen_port or DEFAULT_PORT,
            )

    def config_write(self):
        self.remote.config_write(self._config)

    def config_remove(self):
        self.remote.config_remove()

    def can_peer(self, other: "MeshNode") -> bool:
        """Check if this node can peer with another node.

        Uses the Wireguard UDP listen ports and endpoints to check if the nodes can peer,
        so this can only be reliable when Wireguard is not up on either node.
        """
        print(f"[{self}] [can_peer] {other}", file=sys.stderr)
        return (not self.remote.is_up and self.remote.udping_from(
            listen_port=self.listen_port or DEFAULT_PORT,
            endpoint_host=self.endpoint.host,
            endpoint_port=self.endpoint.port or DEFAULT_PORT,
            remote=other.remote,
        )) or (not other.remote.is_up and other.remote.udping_from(
            listen_port=other.listen_port or DEFAULT_PORT,
            endpoint_host=other.endpoint.host,
            endpoint_port=other.endpoint.port or DEFAULT_PORT,
            remote=self.remote,
        ))

    def as_peer(self, **kwds) -> WireguardPeer:
        return WireguardPeer(
            public_key=self._config.private_key.public_key(),
            allowed_ips=[self.wg_addr],
            endpoint_host=self.endpoint.host,
            endpoint_port=self.endpoint.port or DEFAULT_PORT,
            friendly_name=self.name,
            friendly_json=self.json,
            **kwds,
        )

    def peer_with(self, other: "MeshNode"):
        this_pubkey = self._config.private_key.public_key()
        that_pubkey = other._config.private_key.public_key()

        if this_pubkey == that_pubkey:
            raise ValueError("Cannot peer with self")

        if this_pubkey in other._config.peers and that_pubkey in self._config.peers:
            this_peer = other._config.peers[this_pubkey]
            that_peer = self._config.peers[that_pubkey]

            if this_peer.allowed_ips[0] != self.wg_addr or that_peer.allowed_ips[0] != other.wg_addr:
                raise ValueError("Existing peering WireGuard addresses do not match")

            # noinspection DuplicatedCode
            if not self.json and this_peer.friendly_json is not None:
                self.json = this_peer.friendly_json
            elif self.json and this_peer.friendly_json != self.json:
                this_peer.friendly_json = self.json

            if this_peer.friendly_name != self.name:
                this_peer.friendly_name = self.name

            # noinspection DuplicatedCode
            if not other.json and that_peer.friendly_json is not None:
                other.json = that_peer.friendly_json
            elif other.json and that_peer.friendly_json != other.json:
                that_peer.friendly_json = other.json

            if that_peer.friendly_name != other.name:
                that_peer.friendly_name = other.name

            return

        if not self._mesh.full and not self.can_peer(other):
            return

        preshared_key = WireguardKey(token_bytes(32))
        this_peer = self.as_peer(preshared_key=preshared_key)
        that_peer = other.as_peer(preshared_key=preshared_key)

        self._config.add_peer(that_peer)
        other._config.add_peer(this_peer)

        this_index = self.index
        that_index = other.index
        this_gretap_name = f"gt-{self._mesh.name}{that_index}"
        that_gretap_name = f"gt-{other._mesh.name}{this_index}"
        this_gretap_addr = self.wg_addr.ip
        that_gretap_addr = other.wg_addr.ip
        this_bridge_name = f"br-{self._mesh.name}"
        that_bridge_name = f"br-{other._mesh.name}"
        this_bridge_addr = self.addr
        that_bridge_addr = other.addr

        self._config.postup.extend(gretap_up(
            gretap_name=this_gretap_name,
            bridge_name=this_bridge_name,
            priority=self.bridge_priority,
            local=this_gretap_addr,
            remote=that_gretap_addr,
            bridge_addr=this_bridge_addr,
        ))

        other._config.postup.extend(gretap_up(
            gretap_name=that_gretap_name,
            bridge_name=that_bridge_name,
            priority=other.bridge_priority,
            local=that_gretap_addr,
            remote=this_gretap_addr,
            bridge_addr=that_bridge_addr,
        ))

        self._config.predown.extend(gretap_down(
            gretap_name=this_gretap_name,
            bridge_name=this_bridge_name,
        ))

        other._config.predown.extend(gretap_down(
            gretap_name=that_gretap_name,
            bridge_name=that_bridge_name,
        ))

    def up(self, *, write: bool | None = None) -> bool:
        remote: WireguardRemote = self.remote

        if write or write is None and (write := not remote.config_exists) or (write := remote.config != self._config):
            try:
                remote.config_write(self._config)

            except Exception as exc:
                print(f"[{self}] [up] !! config_write failed: {exc}", file=sys.stderr)
                return False

        if remote.is_up:
            if not write:
                return True
            remote.down()

        if isinstance(up := remote.up(), RuntimeError):
            print(f"[{self}] [up] !! {remote.interface}:\n{up}", file=sys.stderr)
            if write:
                remote.config_remove()
            return False

        else:
            print(f"[{self}] [up] ++ {remote.interface}:\n{up}", file=sys.stderr)

        return True

    def down(self, *, remove: bool | None = None) -> bool:
        remote: WireguardRemote = self.remote

        if remote.is_up:
            if isinstance(down := remote.down(), RuntimeError):
                print(f"[{self}] [down] !! {remote.interface}:\n{down}", file=sys.stderr)

                if not remote.is_up and remove:
                    remote.config_remove()
                return False

            else:
                print(f"[{self}] [down] -- {remote.interface}:\n{down}", file=sys.stderr)

        if remove or remove is None and remote.config_exists:
            remote.config_remove()

        return True

    def sync(self, *, up: bool | None = None) -> bool:
        remote: WireguardRemote = self.remote
        remote_config = remote.config

        if remote_config is None or remote_config != self._config:
            if up or up is None and remote.is_up:
                return self.up(write=True)

            remote.config_write(self._config)
            return True

        return False



from ipaddress import IPv4Interface, IPv6Interface, IPv4Address, IPv6Address


def gretap_up(
        gretap_name: str,
        bridge_name: str,
        priority: int,
        local: IPv4Address | IPv6Address,
        remote: IPv4Address | IPv6Address,
        bridge_addr: IPv4Interface | IPv6Interface,
) -> list[str]:
    six = local.version == 6 and remote.version == 6

    return [
        f"ip link add dev {gretap_name} type {'ip6gretap' if six else 'gretap'} local {local} remote {remote}",
        f"ip link set dev {gretap_name} up",
        (
            f"if [ ! -f /sys/class/net/{bridge_name}/bridge/bridge_id ]; then "
            f"ip link add name {bridge_name} type bridge stp 1 prio {priority}; "
            f"ip link set dev {bridge_name} up; "
            f"ip addr add {bridge_addr} dev {bridge_name}; fi"
        ),
        f"ip link set dev {gretap_name} master {bridge_name}",
    ]


def gretap_down(gretap_name: str, bridge_name: str, *, no_fail: bool = True) -> list[str]:
    nofail = " || true" if no_fail else ""
    return [
        f"ip link set dev {gretap_name} nomaster{nofail}",
        f"ip link del dev {gretap_name}{nofail}",
        (
            f"if ! ip a | grep -q 'master {bridge_name}'; then "
            f"ip link del dev {bridge_name}{nofail}; fi"
        ),
    ]



"""Mesh Dynamic Resource Provider.
"""
import traceback
from typing import Self

import pulumi
from pulumi import Input, Output, ResourceOptions
from pulumi.dynamic import Resource, ResourceProvider, CreateResult, DiffResult, CheckResult, CheckFailure
from pulumi.output import Unknown

from mesh import Mesh

from .util import random_id

__all__ = "WireguardMesh", "MeshArgs", "MeshNodeArgs", "MeshInfo"

MeshInfo = Mesh.Info


@pulumi.input_type
class MeshNodeArgs:
    addr: Input[str]
    name: Input[str] | None = None
    ssh: Input[str] | None = None
    endpoint: Input[str] | None = None
    listen_port: Input[int] | None = None

    def __init__(self, **kwds):
        self.__dict__.update(kwds)

    @classmethod
    def convert(cls, value: dict | Self, *, key: str | None = None) -> Self:
        if isinstance(value, cls):
            return value

        if isinstance(value, dict):
            if key is not None:
                value["name"] = value.setdefault("name", key) or key
            return cls(**value)

        raise TypeError(f"Cannot convert {value!r} to {cls.__name__}")


@pulumi.input_type
class MeshArgs:
    name: Input[str]
    network: Input[str]
    full: Input[bool] | None
    nodes: Input[dict[str, MeshNodeArgs]]

    def __init__(self, **kwds):
        nodes = kwds.pop("nodes")
        self.__dict__.update(kwds)
        self.nodes = {name: MeshNodeArgs.convert(node, key=name) for name, node in nodes.items()}


class WireguardMesh(Resource):
    name: Output[str]
    network: Output[str]
    full: Output[bool] | None
    nodes: Output[dict[str, dict]]
    info: Output[dict]

    def __init__(self, name: str, props: MeshArgs, opts: ResourceOptions | None = None):
        super().__init__(MeshProvider(), name, dict(info=None, **vars(props)), opts)


class MeshProvider(ResourceProvider):
    @classmethod
    def from_props(cls, props: dict) -> Mesh:
        props = props.copy()
        props.pop("info", None)
        props.pop("__provider", None)
        return Mesh(**props)

    def check(self, old: dict, new: dict) -> CheckResult:
        failures = []

        # noinspection PyBroadException
        try:
            if new.get("name") is None:
                failures.append(CheckFailure("name", "required"))

            if new.get("network") is None:
                failures.append(CheckFailure("network", "required"))

            if (nodes := new.get("nodes")) is None:
                failures.append(CheckFailure("nodes", "required"))
            else:
                for name, node in nodes.items():  # type: str, dict

                    if node.get("addr") is None:
                        failures.append(CheckFailure(f"nodes.{name}.addr", "required"))
                    if (name_ := node.get("name")) is not None and name != name_:
                        failures.append(CheckFailure(f"nodes.{name}.name", "must match key"))
                    elif name_ is None:
                        node["name"] = name
                    if node.get("ssh") is None:
                        failures.append(CheckFailure(f"nodes.{name}.ssh", "required"))
                    if node.get("endpoint") is None:
                        failures.append(CheckFailure(f"nodes.{name}.endpoint", "required"))

        except Exception:
            failures.append(CheckFailure(None, "\n" + traceback.format_exc()))  # type: ignore[arg-type]

        return CheckResult(inputs=new, failures=failures)

    def create(self, props: dict) -> CreateResult:
        mesh = self.from_props(props)

        try:
            mesh.peer_all()
            mesh.config_write()

            if not mesh.up(write=False):
                raise RuntimeError("Failed to bring up mesh")

            info = mesh.info
            outs = mesh.asdict(serializer=lambda _, __, v: str(v) if not isinstance(v, int) else v)

        except Exception:
            mesh.down(remove=True)
            raise

        return CreateResult(
            f"mesh-{random_id()}",
            dict(info=info, **outs),
        )

    def diff(self, rid: str, old: dict, new: dict) -> DiffResult:
        replaces = []

        if old["name"] != new["name"]:
            replaces.append("name")

        if old["network"] != new["network"]:
            replaces.append("network")

        if old.get("full") != new.get("full"):
            replaces.append("full")

        old_node_keys = set(old["nodes"].keys())

        for new_node_key, new_node in new["nodes"].items():
            if (old_node := old["nodes"].get(new_node_key)) is None:
                replaces.append(f"nodes.{new_node_key}")
            else:
                old_node_keys.remove(new_node_key)
                for attr in {"addr", "name", "ssh", "endpoint", "listen_port"}:
                    if (ov := old_node.get(attr)) != (nv := new_node.get(attr)):
                        if not isinstance(ov, Unknown) and not isinstance(nv, Unknown):
                            replaces.append(f"nodes.{new_node_key}.{attr}")

        if old_node_keys:
            replaces.extend(f"nodes.{key}" for key in old_node_keys)

        return DiffResult(
            changes=bool(replaces),
            replaces=replaces,
            stables=None,
            delete_before_replace=True,
        )

    def delete(self, rid: str, props: dict) -> None:
        mesh = self.from_props(props)
        mesh.down(remove=True)



from itertools import combinations
from typing import TypedDict

from attrs import define, field

from .types import IPv4Network, IPv6Network, Network, RandomIPv6Address, MeshType, ip_network
from .node import MeshNode


@define(kw_only=True, slots=False, order=True, eq=True)
class Mesh(MeshType):
    nodes: dict[str, MeshNode] = field(converter=MeshNode.convert_all, factory=dict)

    @property
    def pairs(self):
        return combinations(self.nodes.values(), 2)

    @property
    def is_up(self) -> float:
        return sum(int(node.remote.is_up) for node in self.nodes.values()) / len(self.nodes)

    @property
    def config_exists(self) -> float:
        return sum(int(node.remote.config_exists) for node in self.nodes.values()) / len(self.nodes)

    class Info(TypedDict):
        name: str
        network: str
        is_up: float
        config_exists: float
        nodes: dict[str, MeshNode.Info]

    @property
    def info(self) -> Info:
        return self.Info(
            name=self.name,
            network=str(self.network),
            is_up=self.is_up,
            config_exists=self.config_exists,
            nodes={name: node.info for name, node in self.nodes.items()},
        )

    def __attrs_post_init__(self):
        for node in self.nodes.values():
            node.__mesh_post_init__(self)

    def __iter__(self):
        return iter(self.nodes)

    def __getitem__(self, name) -> MeshNode:
        return self.nodes[name]

    def __len__(self):
        return len(self.nodes)

    def __contains__(self, name):
        return name in self.nodes

    def config_write(self) -> None:
        [node.config_write() for node in self.nodes.values()]

    def config_remove(self) -> None:
        [node.config_remove() for node in self.nodes.values()]

    def peer_all(self) -> None:
        for node1, node2 in self.pairs:
            node1.peer_with(node2)

    def up(self, *, write: bool | None = None) -> bool | None:
        up_nodes = []

        if write is not False and any(not node.config_exists or not node.config.peers for node in self.nodes.values()):
            # Some nodes don't have configs or peers yet, so make sure all peer lists are up-to-date.
            for node1, node2 in self.pairs:
                node1.peer_with(node2)

        for node in self.nodes.values():
            if node.up(write=write):
                up_nodes.append(node)
            else:
                for up_node in up_nodes:
                    up_node.down(remove=write)
                return False

        return None if not up_nodes else True

    def down(self, *, remove: bool | None = None) -> bool:
        return all(node.down(remove=remove) for node in self.nodes.values())

    def sync(self, *, up: bool | None = None) -> bool:
        return all(node.sync(up=up) for node in self.nodes.values())

    def show(self) -> None:
        for node in self.nodes.values():
            print(f"{node}\n{node.remote.show()}\n")
